/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef __FLEXIO_OS_MB_H__
#define __FLEXIO_OS_MB_H__

#if defined(__riscv)

/*
 * This API adds write memory barrier to synchronize cpu which is writing to the DMAable memory and DMA device which is reading from the DMAable memory.
 * This API should be used before performing an mmio outbox writes that that initiate a DMA operation from the memory.
 */
#define flexio_os_dma_to_dev_wr_barrier() \
    do { \
        asm volatile("fence ow,ow" ::: "memory"); \
    } while(0)

/*
 * This API adds write memory barrier to synchronize cpu which is writing to the DMAable memory and DMA device reading from the DMAable memory.
 * This API should be used when a valid bit in a DMA buffer indicates that there is some valid data.
 * It is necessary to set the valid bit in DMA buffer before finish writing all the data.
 */
#define flexio_os_dma_to_dev_wr_order_barrier() \
    do { \
        flexio_os_dma_to_dev_wr_barrier(); \
    } while(0)

/*
 * This API adds read memory barrier to synchronize cpu which is reading from the DMAable memory and DMA device writing to the memory.
 */
#define flexio_os_dma_from_dev_rd_barrier() \
    do { \
        asm volatile("fence ir,ir" ::: "memory"); \
    } while(0)

/*
 * This function adds write memory barrier for output registers (window)
 * Use it if you want to be sure that writing to window must be completed before next operation
 */
#define flexio_os_window_write_barrier() \
    do { \
        asm volatile("fence o,o" ::: "memory"); \
    } while (0)

/*
 * This function adds write memory barrier for output registers (window)
 * Use it if you want to be sure that reading from window must be completed before next operation
 */
#define flexio_os_window_read_barrier() \
    do { \
        asm volatile("fence i,i" ::: "memory"); \
    } while (0)

/*
 * This function adds memory barrier for output registers (window) in both direction.
 * Use it before window modification in order to protect that all transactions finished with actual
 * configuration
 */
#define flexio_os_window_rw_barrier() \
    do { \
        asm volatile("fence io,io" ::: "memory"); \
    } while (0)

/*
 * This function adds read and write memory barrier for output registers (window) and for DMA memory.
 * This function provides maximum protection, but it is most slow.
 * Don't use it in performance critical code.
 */
#define flexio_os_all_barrier() \
    do { \
        asm volatile("fence iorw,iorw" ::: "memory"); \
    } while (0)

#else
#error "Unsupported CPU architecture"
#endif

#endif /*__FLEXIO_OS_MB_H__*/
