/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _FLEXIO_SDK_H_
#define _FLEXIO_SDK_H_

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <infiniband/verbs.h>
#include <infiniband/mlx5dv.h>
#include <common/flexio_common.h>

#ifdef __cplusplus
extern "C" {
#endif

#define PHY_CORES_NUM 16
#define HARTS_PER_CORE 16
#define HARTS_LIMIT (PHY_CORES_NUM * HARTS_PER_CORE)
#define OS_RESERVED_HARTS 2
#define USER_HARTS_LIMIT (HARTS_LIMIT - OS_RESERVED_HARTS)
#define HART_BITMASK_LEN (HARTS_LIMIT / 8)

enum {
	FLEXIO_CQ_ELEMENT_TYPE_DPA_THREAD	= 0x0,
	FLEXIO_CQ_ELEMENT_TYPE_EXT_EQ		= 0x2,
	/* value 0x1 reserved */
	FLEXIO_CQ_ELEMENT_TYPE_NON_DPA_CQ	= 0x8,
	FLEXIO_CQ_ELEMENT_TYPE_DPA_MSIX_EMULATED_CQ = 0x9,
};

/* flexio SDK types and structs */

typedef enum {
	FLEXIO_STATUS_SUCCESS	= 0,
	FLEXIO_STATUS_FAILED	= 1,
	FLEXIO_STATUS_TIMEOUT	= 2,
	FLEXIO_STATUS_FATAL_ERR = 3,
} flexio_status;

typedef enum {
	FLEXIO_QPC_ST_RC            = 0x0,
	FLEXIO_QPC_ST_UC            = 0x1,
	FLEXIO_QPC_ST_UD            = 0x2,
	FLEXIO_QPC_ST_XRC           = 0x3,
	FLEXIO_QPC_ST_IBL2          = 0x4,
	FLEXIO_QPC_ST_DCI           = 0x5,
	FLEXIO_QPC_ST_QP0           = 0x7,
	FLEXIO_QPC_ST_QP1           = 0x8,
	FLEXIO_QPC_ST_RAW_DATAGRAM  = 0x9,
	FLEXIO_QPC_ST_REG_UMR       = 0xc,
	FLEXIO_QPC_ST_DC_CNAK       = 0x10,
} flexio_qp_transport_type;

typedef enum {
	FLEXIO_CQE_COMP_NONE	= 0x0,
	FLEXIO_CQE_COMP_BASIC	= 0x1,
	FLEXIO_CQE_COMP_ENH	= 0x2,
} flexio_cqe_comp_type;

typedef enum {
	FLEXIO_CQ_PERIOD_MODE_EVENT	= 0x0,
	FLEXIO_CQ_PERIOD_MODE_CQE	= 0x1,
} flexio_cq_period_mode_t;

struct flexio_process;
struct flexio_window;
struct flexio_outbox;
struct flexio_event_handler;
struct flexio_thread;
struct flexio_cq;
struct flexio_rq;
struct flexio_sq;
struct flexio_qp;
struct flexio_mkey;
struct flexio_uar;
struct flexio_msix;
struct flexio_emu_device_msix_map;
struct flexio_eq_msix_list;
struct flexio_emu_db_to_cq_ctx;
struct flexio_transport_domain;

struct flexio_cq_attr {
	uint8_t log_cq_ring_depth;
	uint8_t element_type;
	union {
		uint32_t *eqn;
		struct flexio_thread *thread;
		struct flexio_msix* msix;
	};
	uint32_t uar_id;
	void *uar_base_addr;
	flexio_cqe_comp_type cqe_comp_type;
	flexio_cq_period_mode_t cq_period_mode;
	uint16_t cq_period;
	uint16_t cq_max_count;
};

struct flexio_wq_attr {
	uint8_t log_wq_ring_depth;
	uint8_t log_data_bsize;
	uint8_t log_wqe_bsize;
	uint32_t uar_id;
	uint32_t user_index;
	struct ibv_pd *pd;
};

struct flexio_process_attr {
	struct ibv_pd *pd;
};

struct flexio_thread_attr {
	const char* func_symbol;
	uint64_t thread_arg;
	flexio_uintptr_t thread_local_storage_daddr;
	uint8_t hart_bitmask[HART_BITMASK_LEN];
};

struct flexio_event_handler_attr {
	const char* func_symbol;
	uint64_t arg;
	flexio_uintptr_t thread_local_storage_daddr;
	uint8_t hart_bitmask[HART_BITMASK_LEN];
};

enum flexio_qp_op_types {
	FLEXIO_QP_WR_RDMA_WRITE= 0x4,
	FLEXIO_QP_WR_RDMA_READ = 0x8,
	FLEXIO_QP_WR_ATOMIC_CMP_AND_SWAP = 0x10,
};

struct flexio_qp_attr {
	uint32_t transport_type;
	uint32_t uar_id;
	uint32_t user_index;
	int qp_access_mask;	/* Expected values: IBV_ACCESS_REMOTE_WRITE, IBV_ACCESS_REMOTE_READ,
				   IBV_ACCESS_REMOTE_ATOMIC, IBV_ACCESS_LOCAL_WRITE. */
	uint32_t ops_flag;	/* values from flexio_qp_op_types */
	struct ibv_pd *pd;

	uint32_t log_sq_buffer_depth;
	int log_data_chunk_bsize;
	uint32_t no_sq;
	uint32_t sq_cqn;

	uint32_t log_rq_buffer_depth;
	uint32_t rq_type;
	uint32_t rq_cqn;
};

struct flexio_qp_set_ready_attr {
	uint32_t remote_qp_num;

	uint8_t gid_table_index;
	uint8_t fl; /* Force Loopback */
	uint8_t *dest_mac;
	union ibv_gid rgid_or_rip;
	uint16_t rlid;
	int qp_access_mask;	/* Expected values: IBV_ACCESS_REMOTE_WRITE, IBV_ACCESS_REMOTE_READ,
				   IBV_ACCESS_REMOTE_ATOMIC, IBV_ACCESS_LOCAL_WRITE. */
	uint32_t min_rnr_nak_timer;
	uint8_t path_mtu;
	uint8_t retry_count;
	uint8_t vhca_port_num;
};

struct flexio_mkey_attr {
	struct ibv_pd *pd;
	flexio_uintptr_t daddr;
	size_t len;
	int access;
};

/* Flex IO API Functions */

/**
 * @brief Allocates a buffer on Flex IO heap memory.
 *
 * This function allocates a buffer with the requested size on the Flex IO heap memory.
 * On success - sets dest_daddr_p to the start address of the allocated buffer.
 * On Failure - sets dest_daddr_p to 0x0.
 *
 * @param[in]  process - A pointer to the Flex IO process context.
 * @param[in]  buff_bsize - The size of the buffer to allocate.
 * @param[out] dest_daddr_p - A pointer to the Flex IO address, where the buffer was allocated.
 *
 * @return flexio status value.
 */
flexio_status flexio_buf_dev_alloc(struct flexio_process *process, size_t buff_bsize,
				   flexio_uintptr_t *dest_daddr_p);

/**
 * @brief Deallocates Flex IO heap memory buffer.
 *
 * This function frees Flex IO heap memory buffer by address.
 *
 * @param[in] process - A pointer to the Flex IO process context.
 * @param[in] daddr - A pointer to an address of allocated memory on the Flex IO heap.
 *
 * @return flexio status value.
 */
flexio_status flexio_buf_dev_free(struct flexio_process *process, flexio_uintptr_t daddr_p);

/**
 * @brief Copy from host memory to a pre-allocted Flex IO heap memory buffer.
 *
 * This function copies data from a buffer on the host memory to a buffer on the Flex IO heap
 * memory.
 *
 * @param[in] process - A pointer to the Flex IO process context.
 * @param[in] src_haddr - An address of the buffer on the host memory.
 * @param[in] buff_bsize - The size of the buffer to copy.
 * @param[in] dest_daddr - Flex IO heap memory buffer address to copy to.
 *
 * @return flexio status value.
 */
flexio_status flexio_host2dev_memcpy(struct flexio_process *process, uintptr_t src_haddr,
				     size_t buff_bsize, flexio_uintptr_t dest_daddr);

/**
 * @brief Sets DPA heap memory buffer to a given value.
 *
 * @param[in] process - A pointer to the Flex IO process context.
 * @param[in] value - A value to set the DPA heap memory buffer to.
 * @param[in] buff_bsize - The size of the Flex IO heap memory buffer.
 * @param[in] dest_daddr - Flex IO heap memory buffer address to set.
 *
 * @return flexio status value.
 */
flexio_status flexio_buf_dev_memset(struct flexio_process *process, int value,
				    size_t buff_bsize, flexio_uintptr_t dest_daddr);

/**
 * @brief Copy from host memory to Flex IO heap memory buffer.
 *
 * This function copies data from a buffer on the host memory to the Flex IO memory.
 * The function allocates memory on the device heap which dest_address points to.
 * It is the caller responsibility to deallocate this memory when it is no longer used.
 *
 * @param[in]  process - A pointer to the Flex IO process context.
 * @param[in]  src_haddr - An address of the buffer on the host memory.
 * @param[in]  buff_bsize - The size of the buffer to copy.
 * @param[out] dest_daddr_p - A pointer to the Flex IO address, where the buffer was copied to.
 *
 * @return flexio status value.
 */
flexio_status flexio_copy_from_host(struct flexio_process *process, uintptr_t src_haddr,
				    size_t buff_bsize, flexio_uintptr_t *dest_daddr_p);

/**
 * @brief Creates a Flex IO CQ.
 *
 * This function creates a Flex IO CQ.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  ibv_ctx - A pointer to an IBV device context (might be different than process').
 * @param[in]  attr - A pointer to the CQ attributes struct.
 * @param[out] cq - A pointer to the created CQ context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_cq_create(struct flexio_process *process, struct ibv_context *ibv_ctx,
			       const struct flexio_cq_attr *attr, struct flexio_cq **cq);

/**
 * @brief Destroys a Flex IO CQ.
 *
 * This function destroys a Flex IO CQ.
 *
 * @param[in] cq - A pointer to a CQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_cq_destroy(struct flexio_cq *cq);

/**
 * @brief Creates an Mkey to the process device UMEM
 *
 * This function creates an MKey over the provided PD for the provided process device
 * UMEM.
 * The mkey_id will point to the field in the containing flexio_mkey object.
 *
 * @param[in]  process - A pointer to the Flex IO process context.
 * @param[in]  fattr - A pointer to a Flex IO MKey attribute struct.
 * @param[out] mkey - A pointer to a pointer to the created MKey struct.
 *
 * @return flexio status value.
 */
flexio_status flexio_device_mkey_create(struct flexio_process *process,
					struct flexio_mkey_attr *fattr, struct flexio_mkey **mkey);

/**
 * @brief destroys an MKey object containing the given ID
 *
 * This function destroys an Mkey object containing the given ID.
 *
 * @param[in] mkey - A pointer to the Flex IO MKey to destroy.
 *
 * @return flexio status value.
 */
flexio_status flexio_device_mkey_destroy(struct flexio_mkey *mkey);

/**
 * @brief Creates a Flex IO UAR object
 *
 * This function creates a Flex IO UAR object from a DevX UAR object.
 *
 * @param[in]  process - A pointer to the Flex IO process context.
 * @param[in]  devx_uar - A pointer to a DevX UAR struct.
 * @param[out] flexio_uar - A pointer to a pointer to the created Flex IO UAR struct.
 *
 * @return flexio status value.
 */
flexio_status flexio_uar_create(struct flexio_process *process, struct mlx5dv_devx_uar *devx_uar,
				struct flexio_uar **flexio_uar);

/**
 * @brief destroys a Flex IO UAR object
 *
 * This function destroys a Flex IO UAR object.
 *
 * @param[in] uar - A pointer to the Flex IO UAR to destroy.
 *
 * @return flexio status value.
 */
flexio_status flexio_uar_destroy(struct flexio_uar *uar);

/**
 * @brief Creates a Flex IO event handler.
 *
 * This function creates a Flex IO event handler for an existing Flex IO process.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  fattr - A pointer to the event handler attributes struct.
 * @param[in]  window_id - Default window ID (or NULL for no default window).
 * @param[in]  outbox_id - Default outbox ID (or NULL for no default outbox).
 * @param[out] event_handler_ptr - A pointer to the created event handler context pointer.
 *
 * @return flexio status value.
 */
flexio_status
flexio_event_handler_create(struct flexio_process *process, struct flexio_event_handler_attr *fattr,
			    const struct flexio_window *window, const struct flexio_outbox *outbox,
			    struct flexio_event_handler **event_handler_ptr);

/**
 * @brief Destroys a Flex IO event handler.
 *
 * This function destroys a Flex IO event handler.
 *
 * @param[in] event_handler - A pointer to an event handler context.
 *
 * @return flexio status value.
 */
flexio_status flexio_event_handler_destroy(struct flexio_event_handler *event_handler);

/**
 * @brief Run a Flex IO event handler.
 *
 * This function makes a Flex IO event handler start running.
 *
 * @param[in] event_handler - A pointer to an event handler context.
 * @param[in] user_arg - A 64 bit argument for the event handler's thread.
 *
 * @return flexio status value.
 */
flexio_status
flexio_event_handler_run(struct flexio_event_handler *event_handler, uint64_t user_arg);

/**
 * @brief Calls a Flex IO process.
 *
 * This function calls a Flex IO process with a function to run from the ELF file.
 * It busy polls until the operation is done on the Flex IO side.
 * This function is a blocking operation and will not resume until a notification
 * from the Flex IO device or a timeout.
 *
 * @param[in]  process - A pointer to the Flex IO process to run.
 * @param[in]  func_symbol - Name of the symbol for the function to run on Flex IO.
 * @param[in]  arg1 - 1st 64 bit argument for the ELF function.
 * @param[in]  arg2 - 2nd 64 bit argument for the ELF function.
 * @param[in]  arg3 - 3rd 64 bit argument for the ELF function.
 * @param[out] func_ret - A pointer to the ELF function return value.
 *
 * @return flexio status value.
 */
flexio_status flexio_process_call(struct flexio_process *process, const char* func_symbol,
				  uint64_t arg1, uint64_t arg2, uint64_t arg3, uint64_t *func_ret);

/**
 * @brief Create a new Flex IO process.
 *
 * This function creates a new Flex IO process with requested image.
 *
 * @param[in]  ibv_ctx - A pointer to a device context.
 * @param[in]  file_start_address - A pointer to the starting location of a the image mapped.
 * @param[in]  file_size - The size of the image.
 * @param[in]  process_attr - Optional, process attributes for create. Can be NULL.
 * @param[out] process_ptr - A pointer to the created process pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_process_create(struct ibv_context *ibv_ctx, uint64_t *file_start_address,
				    size_t file_size,
				    const struct flexio_process_attr *process_attr,
				    struct flexio_process **process_ptr);

/**
 * @brief Destroys a new Flex IO process.
 *
 * This function destroys a new Flex IO process.
 *
 * @param[in] process - A pointer to a process.
 *
 * @return flexio status value.
 */
flexio_status flexio_process_destroy(struct flexio_process *process);

/**
 * @brief Creates a Flex IO RQ.
 *
 * This function creates a Flex IO RQ.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  ibv_ctx - A pointer to an IBV device context (might be different than process').
 * @param[in]  cq - A pointer to a Flex IO CQ.
 * @param[in]  fattr - A pointer to the RQ attributes struct.
 * @param[out] rq - A pointer to the created RQ context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_rq_create(struct flexio_process *process, struct ibv_context *ibv_ctx,
			       const struct flexio_hw_cq *cq, const struct flexio_wq_attr *fattr,
			       struct flexio_rq **flexio_rq_ptr);

/**
 * @brief Destroys a Flex IO RQ.
 *
 * This function destroys a Flex IO RQ.
 *
 * @param[in] rq - A pointer to an RQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_rq_destroy(struct flexio_rq *flexio_rq);

/**
 * @brief Creates a Flex IO SQ.
 *
 * This function creates a Flex IO SQ.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  ibv_ctx - A pointer to an IBV device context (might be different than process').
 * @param[in]  cq_num - A CQ number (can be Flex IO or host CQ).
 * @param[in]  fattr - A pointer to the SQ attributes struct.
 * @param[out] sq - A pointer to the created SQ context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_sq_create(struct flexio_process *process, struct ibv_context *ibv_ctx,
			       uint32_t cq_num, const struct flexio_wq_attr *fattr,
			       struct flexio_sq **flexio_sq_ptr);

/**
 * @brief Destroys a Flex IO SQ.
 *
 * This function destroys a Flex IO SQ.
 *
 * @param[in] sq - A pointer to an SQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_sq_destroy(struct flexio_sq *flexio_sq);

/**
 * @brief Creates and starts a Flex IO thread.
 *
 * This function Creates a Flex IO thread with the given function to run and starts it.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  fattr - User level thread creation attributes.
 * @param[out] thread - A pointer to the created thread context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_thread_create(struct flexio_process *process, struct flexio_thread_attr *fattr,
				   struct flexio_thread **thread);

/**
 * @brief Destroys a Flex IO thread.
 *
 * This function destroys a Flex IO thread.
 *
 * @param[in] thread - A pointer to a thread context.
 *
 * @return flexio status value.
 */
flexio_status flexio_thread_destroy(struct flexio_thread *thread);

/**
 * @brief Creates a Flex IO window.
 *
 * This function Creates a Flex IO window for the given process.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  pd - A pointer to a protection domain struct to the memory the window should access.
 * @param[out] window - A pointer to the created window context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_window_create(struct flexio_process *process, struct ibv_pd *pd,
				   struct flexio_window **window);

/**
 * @brief Destroys a Flex IO window.
 *
 * This function destroys a Flex IO window.
 *
 * @param[in] window - A pointer to a window context.
 *
 * @return flexio status value.
 */
flexio_status flexio_window_destroy(struct flexio_window *window);

/**
 * @brief Creates a Flex IO outbox.
 *
 * This function Creates a Flex IO outbox for the given process.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  other_ctx - An IBV context for creating the PRM object (Different than process's on
 *                         multi GVMI case, NULL or same as process's otherwise).
 * @param[in]  uar - A Flex IO UAR struct created for the outbox's IBV device.
 * @param[out] outbox - A pointer to the created outbox context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_outbox_create(struct flexio_process *process, struct ibv_context *other_ctx,
				   struct flexio_uar *uar, struct flexio_outbox **outbox);

/**
 * @brief Destroys a Flex IO outbox.
 *
 * This function destroys a Flex IO outbox.
 *
 * @param[in] outbox - A pointer to a outbox context.
 *
 * @return flexio status value.
 */
flexio_status flexio_outbox_destroy(struct flexio_outbox *outbox);

/**
 * @brief Create environment to support print from DPA.
 *
 * This function allocates resources to support printing from Flex IO to HOST.
 *
 * Print works in synchronous or asynchronous modes.
 * In case of synchronous mode dedicated thread started to receive data and print it immediately.
 * In case of asynchronous mode all print buffer will be flushed by flexio_print_flush().
 * Buffer can be overrun.
 *
 * This function doesn't have "destroy" pair. All printing infrastructure will be closed
 * and resources will be released from flexio_process_destroy() function.
 *
 * @param[in] process - A pointer to the Flex IO process.
 * @param[in] flexio_uar - A pointer to a Flex IO UAR object created by caller for device side
 * @param[in] data_bsize - size of buffer, used for data transfer from Flex IO to HOST
 * 			  MUST be power of two and be at least 2Kb
 * @param[in] out - file to save data from Flex IO. Use stdout if you want receive data
 * 		    on HOST's console
 * @param[in] is_async - select between sync(recommended)/async modes
 * @param[out] ppthread - A pointer to receive phread ID of created thread. May be NULL if
 * 			  user doesn't need it.
 *
 * @return flexio status value.
 */
flexio_status flexio_print_init(struct flexio_process *process, struct flexio_uar *flexio_uar,
				size_t data_bsize, FILE *out, int is_async, pthread_t *ppthread);

/**
 * @brief Flush print buffer in case of asynchronous print.
 *
 * All data from print buffer will be flushed to file, definded in flexio_print_init().
 *
 * In case of synchronous print this functions does nothing.
 * This function allocates resources to support printing from Flex IO to HOST.
 *
 * @param[in] process - A pointer to the Flex IO process.
 *
 * @return flexio status value.
 */
flexio_status flexio_print_flush(struct flexio_process *process);

/**
 * @brief Get file descriptor for error handler
 *
 * User should get fd in order to monitor for nonrecoverable errors
 *
 * User can poll all created processes, using select/poll/epoll
 * functions family.
 *
 * @param[in] process - A pointer to the Flex IO process.
 *
 * @return - file descriptor.
 */
int flexio_err_handler_fd(struct flexio_process *process);

/**
 * @brief Check if unrecoverable error occurred
 *
 * User can call this function many times. Error status will be reset only after calling
 * flexio_err_display();
 *
 * It is suggested to check error status after every negotiation with DPA and periodically later.
 * Alternative way - to poll file descriptor got from flexio_err_handler_fd() in order to get
 * events about error
 *
 * @param[in] process - A pointer to the Flex IO process.
 *
 * @return - nonzero value if error happen.
 */
int flexio_err_status(struct flexio_process *process);

/**
 * @brief Print error data to stderr
 *
 * User should call this function, once he got information, that an unrecoverable error in DPA has
 * happened.
 *
 * Calling this function will reset error status.
 *
 * @param[in] process - A pointer to the Flex IO process.
 */
void flexio_err_display(struct flexio_process *process);

#define FLEXIO_MAX_EQ_TO_MSIX_MAP_PAIRS 64
/**
 * @brief Query mapping of EQs to MSI-X vectors.
 *
 * This function gets emulated devices VHCA ID and return mapping of EQs to MSI-X vectors
 * for this emulated device.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  emualterd_device_vhca_id - An emulated device VHCA ID.
 * @param[in]  max_msix - maximal msix to query, should not exceed FLEXIO_MAX_EQ_TO_MSIX_MAP_PAIRS
 * @param[out] map - A pointer to receive list of EQs to MSI-X vectors mapping.
 *
 * @return flexio status value.
 */
flexio_status flexio_emulated_device_msix_map_create(struct flexio_process *process,
						     uint16_t emulated_device_vhca_id,
						     uint16_t max_msix,
						     struct flexio_emu_device_msix_map** map);

/**
 * @brief Free list of EQs to MSI-X vectors mapping.
 *
 * This function frees a list of EQs to MSI-X vectors mapping.
 *
 * @param[in] map - A point to list of EQs to MSI-X vectors mapping.
 *
 * @return flexio status value.
 */
flexio_status flexio_emulated_device_msix_map_free(struct flexio_emu_device_msix_map* map);

/**
 * @brief Mapping emulated devices doorbell to CQ.
 *
 * This function maps emulated device doorbell to Flex IO CQ.
 * A doorbell from the emulated device queue will create CQE on the specified CQ.
 *
 * @param[in]  ibv_ctx - A pointer to a device context.
 * @param[in]  emu_dev_vhca_id - An emulated device VHCA ID.
 * @param[in]  emu_dev_queue_id - A queue of the emulated device.
 * @param[in]  cq - A pointer to the Flex IO CQ.
 * @param[out] emu_db_to_cq_ctx - A pointer to the created emulated device doorbell to CQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_emu_db_to_cq_map(struct ibv_context *ibv_ctx, uint32_t emu_dev_vhca_id,
				      uint32_t emu_dev_queue_id, struct flexio_cq *cq,
				      struct flexio_emu_db_to_cq_ctx** emu_db_to_cq_ctx);

/**
 * @brief Unmapping emulated devices doorbell to CQ.
 *
 * This function unmaps emulated device queue doorbell to Flex IO CQ.
 *
 * @param[in] emu_device_to_cq_ctx - A pointer to emulated device doorbell to CQ context.
 *
 * @return flexio status value.
 */
flexio_status flexio_emu_db_to_cq_unmap(struct flexio_emu_db_to_cq_ctx* emu_db_to_cq_ctx);

/**
 * @brief Create emulated device msix.
 *
 * This function creates emulated device msix infrastructure.
 *
 * @param[in] process - A pointer to the emulation manager Flex IO process.
 * @param[in] target_ibv_ctx - A pointer to a SF device context.
 * @param[in] direct_mode - a flag to indicate EQ is to be used directly by app.
 * @param[in] target_uar_id - uar page id to be used for emulated cq created, relevant when
 *                            direct_mode is false.
 * @param[in] msix_id - The msix id.
 * @param[in] map - A pointer to msix  to eq mapping.
 * @param[out] msix - A pointer to receive the flexio_msix handle.
 *
 * @return flexio status value.
 */
flexio_status flexio_emulated_device_msix_create(struct flexio_process *process,
						 struct ibv_context *target_ibv_ctx,
						 uint8_t direct_mode, uint32_t target_uar_id,
						 uint16_t msix_id,
						 struct flexio_emu_device_msix_map* map,
						 struct flexio_msix** msix);

/**
 * @brief destroy emulated device msix.
 *
 * This function destroy emulated device msix infrastructure.
 *
 * @param[in] msix - A pointer to the flexio_msix handle structure.
 *
 * @return flexio status value.
 */
flexio_status flexio_emulated_device_msix_destroy(struct flexio_msix* msix);

/**
 * @brief Creates a Flex IO QP.
 *
 * This function creates a Flex IO QP.
 *
 * @param[in]  process - A pointer to the Flex IO process.
 * @param[in]  ibv_ctx - A pointer to an IBV device context (might be different than process').
 * @param[in]  qp_fattr - A pointer to the QP attributes struct.
 * @param[out] qp_ptr - A pointer to the created QP context pointer.
 *
 * @return flexio status value.
 */
flexio_status flexio_qp_create(struct flexio_process *process, struct ibv_context *ibv_ctx,
			       struct flexio_qp_attr *qp_fattr, struct flexio_qp **qp_ptr);

/**
 * @brief set Flex IO QP to RTS state.
 *
 * This function modifies Flex IO QP state from rst->init, and from init->rtr and then rtr->rts.
 * At the end of the procedure Flex IO qp would have moved from rst to rts state.
 *
 * @param[in] qp - A pointer to the QP context.
 * @param[in] fattr - A pointer to the attributes struct that will define the QP connection.
 *
 * @return flexio status value.
 */
flexio_status flexio_qp_set_ready(struct flexio_qp *qp, struct flexio_qp_set_ready_attr* fattr);

/**
 * @brief Destroys a Flex IO QP.
 *
 * This function destroys a Flex IO QP.
 *
 * @param[in] qp - A pointer to the QP context.
 *
 * @return flexio status value.
 */
flexio_status flexio_qp_destroy(struct flexio_qp *qp);

/**
 * @brief Set the bit in the given mask corresponding to the given core and hart
 *
 * This function calculates the bit to set in the hart bitmask for the given core and hart.
 *
 * @param[in]  process - A pointer to the process context.
 * @param[in]  hart - hart global number to use (in range 0..(USER_HARTS_LIMIT-1) )
 * @param[out] mask - HART bitmask array to set (size HART_BITMASK_LEN).
 *
 * @return flexio status value.
 */
flexio_status flexio_hart_mask_bit_set(struct flexio_process *process, int hart, uint8_t mask[]);

/**
 * @brief Reset all bits of given mask for hart selection
 *
 * This function clears all bits. May be useful in case of reuse of attributes.
 *
 * @param[out] mask - HART bitmask array to reset (size HART_BITMASK_LEN).
 *
 * @return flexio status value.
 */
void flexio_hart_mask_reset(uint8_t mask[]);

/* Helpers and Getters */
struct flexio_thread *flexio_event_handler_get_thread(struct flexio_event_handler *event_handler);
uint32_t flexio_event_handler_get_id(struct flexio_event_handler *event_handler);
struct flexio_hw_cq *flexio_cq_get_hw_cq(struct flexio_cq *cq);
struct flexio_hw_wq *flexio_rq_get_hw_rq(struct flexio_rq *rq);
struct mlx5dv_devx_obj *flexio_rq_get_tir(struct flexio_rq *rq);
struct flexio_hw_wq *flexio_sq_get_hw_sq(struct flexio_sq *sq);
struct flexio_hw_qp *flexio_qp_get_hw_qp(struct flexio_qp *qp);
uint32_t flexio_outbox_get_id(struct flexio_outbox *outbox);
uint32_t flexio_window_get_id(struct flexio_window *window);
uint32_t flexio_emu_db_to_cq_ctx_get_id(struct flexio_emu_db_to_cq_ctx *emu_db_to_cq_ctx);
uint32_t flexio_emulated_device_msix_get_eqn(struct flexio_msix *msix);
uint32_t flexio_emulated_device_msix_get_cqn(struct flexio_msix *msix);
uint32_t flexio_mkey_get_id(struct flexio_mkey *mkey);

#define HART_NUM(core_num, hart_num_in_core) ((core_num) * HARTS_PER_CORE + hart_num_in_core)

/*
 * Set multiple harts in one command with variadic list of arguments
 * Usage examples (first and second example provide the same result):
 *
 * SET_HART(mask, HART_NUM(0, 0), HART_NUM(0, 1), HART_NUM(15, 13));
 * SET_HART(mask, 0, 16, 253);
 */
#define SET_HART(mask, ... ) calc_harts_mask(mask, __VA_ARGS__, -1)
void set_harts_mask(uint8_t mask[], ...);

#ifdef __cplusplus
}
#endif
#endif /* _FLEXIO_SDK_H_ */
