/*
 * SPDX-FileCopyrightText: Copyright (c) 2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-3-Clause
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 *
 * 3. Neither the name of the copyright holder nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef _FLEXIO_ELF_H_
#define _FLEXIO_ELF_H_

#include <stdlib.h>
#include <stdio.h>
#include <libflexio/flexio.h>

#ifdef __cplusplus
extern "C" {
#endif

/**
 * @brief Extract device ELF embedded in the input ELF file.
 *
 * This function extract a device ELF embedded in a dedicated section in
 * the input ELF file, typically host ELF file
 *
 * @param[in]  file_name - Input ELF file name.
 * @param[out] elf_buf - A pointer to a buffer with device ELF content.
 * @param[out] elf_size - The size of the elf buffer.
 *
 * @return flexio status value.
 */
flexio_status flexio_extract_emb_dev_elf(const char *file_name, void **elf_buf, size_t *elf_size);

/**
 * @brief Get the ELF file content.
 *
 * This function opens the elf file, validates it and returns a buffer with
 * ELF content and size.
 *
 * @param[in]  file_name - ELF file name.
 * @param[out] elf_buf - A pointer to a buffer with ELF file content.
 * @param[out] buf_size - The size of the elf buffer.
 *
 * @return flexio status value.
 */
flexio_status flexio_get_elf_file(const char *file_name, void **elf_buf, size_t *buf_size);

/**
 * @brief Get the address of a device function in a DPA process.
 *
 * This function returns value of the symbol representing the address of the device
 * function in the DPA process.
 *
 * @param[in] process - A pointer to the Flex IO process.
 * @param[in] symbol - A symbol representing the device function.
 *
 * @return device function address, in case symbol not found returns NULL
 */
flexio_uintptr_t flexio_get_process_symbol_address(const struct flexio_process *process,
						   const char *symbol);

#ifdef __cplusplus
}
#endif
#endif /* _FLEXIO_ELF_H_ */
